import moment from "moment";
import { DatabaseError } from "./error";
import { TSVector } from "./type";
import { escapeID, escapeValue } from "./util";
import type { Class, Decorators, IDLikeFieldName } from "./types";

export interface IEntityFieldConfig {
	/** 字段名 */
	name: string
	/** 解析后的属性名 */
	prop: string
	/** 是否是主键 */
	primary: boolean
	/** 是否是虚拟字段 */
	virtual: boolean
	/** 编码方式(存入数据库时) */
	encode: (data: any) => string
	/** 解码方式(查询数据库时) */
	decode: (data: any) => any
}


export interface IEntityJoinConfig {
	/** 属性名称 */
	prop: string
	/** 是否是many */
	many: boolean
	/** 目标表 */
	entity: () => any
	/**
	 * join条件生成方法
	 * @param cur 当前表名
	 * @param dst 目标表名
	 */
	builder(cur: string, dst: string): string
}

export interface IEntityConfig {
	/** 类信息 */
	clazz: any
	/** 模式名 */
	schema?: string
	/** 表名 */
	table: string
	/** 字段配置 */
	fields: IEntityFieldConfig[]
	/** join信息 */
	joins: IEntityJoinConfig[]
}

export class BasicEntity {
	private ___is_db_entity___ = true

	public static make<T extends BasicEntity>(this: Class<T>, data: Partial<Pick<T, BasicFieldName<T>>>, prefix?: string) {
		const obj = new this();
		const config = getEntityConfig(this);
		for (const col of config.fields) {
			const pname = prefix ? `${prefix}${col.prop}` : col.prop;
			if (pname in data) (obj as any)[col.prop] = (data as any)[pname];
		}
		return obj;
	}

	/**
	 * 提取指定属性
	 * @param keys 要提取的属性
	 */
	public pick<K extends BasicFieldName<this>>(...keys: K[]): Pick<this, K> {
		const result: any = {};
		for (const k of keys) {
			let val: any = this[k];
			if (typeof val?.toPickData === "function") val = val.toPickData();
			result[k] = val;
		}
		return result;
	}

	/**
	 * 排除指定的属性，提取剩余属性
	 * @param keys 要排除的键
	 */
	public omit<K extends BasicFieldName<this>>(...keys: K[]): Omit<Pick<this, BasicFieldName<this>>, K> {
		const result: any = {};
		const config = getEntityConfigWithoutCheck(this.constructor)!;
		for (const col of config.fields) {
			if (keys.includes(col.prop as K)) continue;
			let val: any = this[col.prop as K];
			if (typeof val?.toPickData === "function") val = val.toPickData();
			result[col.prop as K] = val;
		}
		return result;
	}
}

function entityConfigOf(target: any): IEntityConfig {
	const cnf = target.__entity_config__ as IEntityConfig;

	if (!cnf) target.__entity_config__ = {
		clazz: target,
		fields: [],
		joins: []
	};
	else if (cnf.clazz != target) target.__entity_config__ = {
		...cnf,
		clazz: target,
		fields: [...cnf.fields],
		joins: [...cnf.joins],
	}
	return target.__entity_config__;
}

function getEntityConfigWithoutCheck(target: any): IEntityConfig | null {
	return target?.__entity_config__ ?? null;
}


/**
 * 指定实体的表名
 * @param table 表名，如果需要指定schema，可以使用"<schema>.<table>"的格式
 */
export function Table(table: string): Decorators.ClassDecorator<BasicEntity> {
	return function (target) {
		const [schema, tbl] = table.split(".").map(s => s.trim()).filter(s => !!s);
		const conf = entityConfigOf(target);
		conf.table = table;
		if (!tbl) {
			conf.schema = undefined;
			conf.table = schema;
		}
		else {
			conf.schema = schema;
			conf.table = tbl;
		}
	}
}

/**
 * 定义一个字段
 * @param name 字段名称
 * @param encode 字段编码凡是
 * @param decode 字段解码方式
 * @param primary 是否是主键
 * @param virtual 是否是虚拟字段
 */
export function Field(name: string | undefined, encode: IEntityFieldConfig["encode"], decode: IEntityFieldConfig["decode"], primary: boolean, virtual: boolean): Decorators.PropDecorator<BasicEntity> {
	return function (target, propertyKey) {
		const conf = entityConfigOf(target.constructor)
		conf.fields.push({ name: name ?? propertyKey, prop: propertyKey, encode, decode, primary, virtual });
	}
}


/**
 * 定义一个整形字段
 * @param name 字段名称
 * @param option 字段选项
 */
export function IntColumn(name: string, option?: { primary: boolean, virtual?: boolean }): Decorators.PropDecorator<BasicEntity>
export function IntColumn(option?: { primary: boolean, virtual?: boolean }): Decorators.PropDecorator<BasicEntity>
export function IntColumn(name?: any, option?: any) {
	if (typeof name != "string") {
		option = name;
		name = undefined;
	}
	return Field(name, v => escapeValue(v), v => v, option?.primary ?? false, option?.virtual ?? false);
}


/**
 * 定义浮点数(包括定点小数)
 * @param name 字段名称
 * @param option 字段选项
 */
export function FloatColumn(option?: { virtual?: boolean }): Decorators.PropDecorator<BasicEntity>
export function FloatColumn(name: string, option?: { virtual?: boolean }): Decorators.PropDecorator<BasicEntity>
export function FloatColumn(name?: any, option?: any) {
	if (typeof name !== "string") {
		option = name;
		name = undefined;
	}

	return Field(name, v => escapeValue(v), v => parseFloat(v), false, option?.virtual ?? false);
}


/**
 * 定义字符串字段(支持，char varchar text)
 * @param name 字段名称
 * @param option 字段选项
 */
export function StringColumn(option?: { primary?: boolean, virtual?: boolean }): Decorators.PropDecorator<BasicEntity>
export function StringColumn(name: string, option?: { primary?: boolean, virtual?: boolean }): Decorators.PropDecorator<BasicEntity>
export function StringColumn(name?: any, option?: any) {
	if (typeof name != "string") {
		option = name
		name = undefined
	}
	return Field(name, v => escapeValue(v), v => v, option?.primary ?? false, option?.virtual ?? false);
}

/**
 * 定义布尔字段
 * @param name 字段名称
 * @param option 字段选项
 */
export function BooleanColumn(option?: { virtual?: boolean }): Decorators.PropDecorator<BasicEntity>
export function BooleanColumn(name: string, option?: { virtual?: boolean }): Decorators.PropDecorator<BasicEntity>
export function BooleanColumn(name?: any, option?: any) {
	if (typeof name != "string") {
		option = name
		name = undefined
	}

	const checker = (v: any) => {
		if (typeof v == "boolean") return v;
		if (typeof v != "string") return false;
		v = v.toLowerCase();
		return ["true", "yes", "on"].includes(v)
	}

	return Field(name, (v) => escapeValue(checker(v)), checker, false, option?.virtual ?? false);
}

/**
 * 定义日期字段
 * @param name 字段名称
 * @param option 字段选项
 */
export function DateColumn(option?: { primary?: boolean, virtual?: boolean }): Decorators.PropDecorator<BasicEntity>
export function DateColumn(name: string, option?: { primary?: boolean, virtual?: boolean }): Decorators.PropDecorator<BasicEntity>
export function DateColumn(name?: any, option?: any) {
	if (typeof name != "string") {
		option = name
		name = undefined
	}

	return Field(name, v => {
		if (v instanceof DateColumn) v = moment(v);
		if (typeof v == "string") v = moment(v);
		if (!moment.isMoment(v)) return escapeValue(null);
		return escapeValue(v.format("YYYY-MM-DD"));
	}, v => {
		if (typeof v == "string" || (v instanceof global.Date)) return moment(v).format("YYYY-MM-DD");
		return null;
	}, option?.primary ?? false, option?.virtual ?? false);
}

/**
 * 定义日期时间字段
 * @param name 字段名称
 */
export function DateTimeColumn(option?: { primary?: boolean, virtual?: boolean }): Decorators.PropDecorator<BasicEntity>
export function DateTimeColumn(name: string, option?: { primary?: boolean, virtual?: boolean }): Decorators.PropDecorator<BasicEntity>
export function DateTimeColumn(name?: any, option?: any) {
	if (typeof name != "string") {
		option = name
		name = undefined
	}

	return Field(name, v => {
		if (v instanceof DateColumn) v = moment(v);
		if (typeof v == "string") v = moment(v);
		if (!moment.isMoment(v)) return escapeValue(null);
		return escapeValue(v.format("YYYY-MM-DD HH:mm:ss"));
	}, v => {
		if (typeof v == "string" || (v instanceof global.Date)) return moment(v).format("YYYY-MM-DD HH:mm:ss");
		return null;
	}, option?.primary ?? false, option?.virtual ?? false);
}


//数组定义
function _Array<T>(name: string | undefined, typeName: string, itemEncoder: (v: any) => (string | null), itemDecoder: ((v: any) => (T | null)) | undefined, virtual: boolean) {
	return Field(name, v => {
		if (!(v instanceof Array)) throw new DatabaseError("DB_ENTITY_FIELD_NOT_ARRAY", `字段${name}需要给定一个数组`);
		return escapeValue(`{${v.map(item => itemEncoder(item)).join(",")}}`) + `::${typeName}`;
	}, v => {
		if (!(v instanceof Array)) return null;
		if (itemDecoder) return v.map(item => itemDecoder(item));
		else return v;
	}, false, virtual);
}

/**
 * 定义整数数组，可以通过option.bytes来设置整数长度，默认为4
 * @param name 字段名称
 * @param option 字段选项
 */
export function IntArrayColumn(name: string, option?: { bytes?: 2 | 4 | 8, virtual?: boolean }): Decorators.PropDecorator<BasicEntity>
export function IntArrayColumn(option?: { bytes?: 2 | 4 | 8, virtual?: boolean }): Decorators.PropDecorator<BasicEntity>
export function IntArrayColumn(name?: any, option?: any) {
	if (name && typeof name !== "string") {
		option = name
		name = undefined
	}

	return _Array(name, `int${option?.bytes ?? 4}[]`, item => {
		item = parseInt(item);
		if (isNaN(item) || !isFinite(item)) return null;
		return item.toString();
	}, item => {
		item = parseInt(item);
		if (isNaN(item) || !isFinite(item)) return null;
		return item;
	}, option?.virtual ?? false);
}

/**
 * 定义字符串数组
 * @param name 字段名称
 * @param option 字段选项
 */
export function StringArrayColumn(option?: { virtual?: boolean }): Decorators.PropDecorator<BasicEntity>
export function StringArrayColumn(name: string, option?: { virtual?: boolean }): Decorators.PropDecorator<BasicEntity>
export function StringArrayColumn(name?: any, option?: any) {
	if (typeof name != "string") {
		option = name
		name = undefined
	}

	return _Array(name, "varchar[]", item => {
		if (typeof item == "string") return item;
		else return null;
	}, item => item, option?.virtual ?? false);
}

/**
 * 定义jsonb字段
 * @param name 字段名称
 */
export function JsonColumn(option?: { virtual?: boolean }): Decorators.PropDecorator<BasicEntity>
export function JsonColumn(name: string, option?: { virtual?: boolean }): Decorators.PropDecorator<BasicEntity>
export function JsonColumn(name?: any, option?: any) {
	if (typeof name != "string") {
		option = name
		name = undefined
	}

	return Field(name, v => escapeValue(JSON.stringify(v)) + "::jsonb", v => v, false, option?.virtual ?? false);
}

/**
 * 定义tsvector字段
 * @param name 字段名称
 */
export function TSVectorColumn(name?: string) {
	return Field(name, v => {
		if (!(v instanceof TSVector)) throw new DatabaseError("DB_ENTITY_FIELD_NOT_TSVECTOR", `字段${name}需要给定一个TSVector`)
		return `E'${v.value}'::tsvector`;
	}, v => TSVector.fromValue(v), false, false);
}

//获取实体配置
export function getEntityConfig(clazz: any) {
	const conf = getEntityConfigWithoutCheck(clazz);
	if (!conf) throw new DatabaseError("DB_ENTITY_NOT_DECLARE", `实体${clazz.name}未注册`);
	return conf;
}

//获取表名
export function getTableName(clazz: any) {
	const tname = getEntityConfigWithoutCheck(clazz)?.table;
	if (!tname) throw new DatabaseError("DB_ENTITY_TABLE_NOT_DECLARE", `实体${clazz.name}缺少表格装饰`);
	return tname;
}

//获取字段配置
export function getField(clazz: any, prop: string) {
	const field = getEntityConfigWithoutCheck(clazz)?.fields.find(f => f.prop == prop);
	if (!field) throw new DatabaseError("DB_ENTITY_FIELD_NOT_DECLARE", `实体${clazz.name}缺少字段${prop}的装饰`);
	return field;
}

//获取字段名
export function getFieldName(clazz: any, prop: string) {
	return getField(clazz, prop).name;
}

//获取主键字段名
export function getPrimaryFieldName(clazz: any) {
	const pname = getEntityConfigWithoutCheck(clazz)?.fields.find(f => f.primary)?.name;
	if (!pname) throw new DatabaseError("DB_ENTITY_FIELD_NOT_DECLARE", `实体${clazz.name}缺少主键装饰`);
	return pname;
}

export function getJoin(clazz: any, prop: string) {
	const join = getEntityConfigWithoutCheck(clazz)?.joins.find(j => j.prop == prop);
	if (!join) throw new DatabaseError("DB_ENTITY_JOIN_NOT_DECLARE", `实体${clazz.name}的${prop}属性缺少join装饰`);
	return join;
}

/**
 * 一对一或多对一Join
 * @param entity 要Join的实体
 * @param propOrBuilder Join的字段或Join构造器
 */
export function JoinOne<T extends BasicEntity, E extends BasicEntity>(entity: () => Class<E>, prop: IDLikeFieldName<T>): Decorators.PropDecorator<T>
export function JoinOne<T extends BasicEntity, E extends BasicEntity>(entity: () => Class<E>, prop: IDLikeFieldName<T>, ref: IDLikeFieldName<E>): Decorators.PropDecorator<T>
export function JoinOne<T extends BasicEntity, E extends BasicEntity>(entity: () => Class<E>, builder: ((cur: string, dst: string) => string)): Decorators.PropDecorator<T>
export function JoinOne<T extends BasicEntity, E extends BasicEntity>(entity: () => Class<E>, propOrBuilder: IDLikeFieldName<T> | ((cur: string, dst: string) => string), ref?: IDLikeFieldName<E>): Decorators.PropDecorator<T> {
	return function (target, propertyKey) {
		const conf = entityConfigOf(target.constructor);
		conf.joins.push({
			prop: propertyKey,
			entity: entity,
			many: false,
			builder: (cur, dst) => {
				if (typeof propOrBuilder == "function") return propOrBuilder(cur, dst);
				const refName = ref ? getFieldName(entity(), ref as string) : getPrimaryFieldName(entity());
				// return `"${cur}"."${getFieldName(target.constructor, propOrBuilder as string)}"="${dst}"."${getPrimaryFieldName(entity())}"`
				return `${escapeID(cur, getFieldName(target.constructor, propOrBuilder as string))}=${escapeID(dst, refName)}`
			}
		});
	}
}


/**
 * 一对多Join
 * @param entity 要Join的实体
 * @param prop Join的字段
 * @param ref 当前表引用的字段
 * @param builder Join构造器
 */
export function JoinMany<T extends BasicEntity, E extends BasicEntity>(entity: () => Class<E>, prop: IDLikeFieldName<E>): Decorators.PropDecorator<T>
export function JoinMany<T extends BasicEntity, E extends BasicEntity>(entity: () => Class<E>, prop: IDLikeFieldName<E>, ref: IDLikeFieldName<T>): Decorators.PropDecorator<T>
export function JoinMany<T extends BasicEntity, E extends BasicEntity>(entity: () => Class<E>, builder: ((cur: string, dst: string) => string)): Decorators.PropDecorator<T>
export function JoinMany<T extends BasicEntity, E extends BasicEntity>(entity: () => Class<E>, propOrBuilder: IDLikeFieldName<E> | ((cur: string, dst: string) => string), ref?: IDLikeFieldName<T>): Decorators.PropDecorator<T> {
	return function (target, propertyKey) {
		const conf = entityConfigOf(target.constructor);
		conf.joins.push({
			prop: propertyKey,
			entity: entity,
			many: true,
			builder: (cur, dst) => {
				if (typeof propOrBuilder == "function") return propOrBuilder(cur, dst);
				const refName = ref ? getFieldName(target.constructor, ref as string) : getPrimaryFieldName(target.constructor);
				// return `"${dst}"."${getFieldName(entity(), propOrBuilder as string)}"="${cur}"."${getPrimaryFieldName(target.constructor)}"`
				return `${escapeID(dst, getFieldName(entity(), propOrBuilder as string))}=${escapeID(cur, refName)}`
			}
		});
	}
}

