import { QueryResult, QueryResultRow } from "pg";
import { BasicEntity, IEntityFieldConfig, IEntityJoinConfig, getEntityConfig, getField, getJoin, getTableName } from "./entity";
import type { JoinaFieldName, Class, ArrayTypeSelf, Values, StringArgs, SQLStringArgs } from "./types";
import { escapeID, escapeValue, formatSQL } from "./util";


//编译器原因，不加此段代码会出错，具体原因未知，请保持定义和types.ts重的同名定义一致
/** 取基础类型的字段名 */
export type BasicFieldName<T extends BasicEntity> = Values<{ [P in keyof T]:
	//忽略never
	Exclude<T[P], null | undefined> extends never ? never : (
		//忽略函数
		Exclude<T[P], null | undefined> extends Function ? never : (
			//忽略实体
			ArrayTypeSelf<Exclude<T[P], null | undefined>> extends BasicEntity ? never : (
				P
			)
		)
	)
}>;

interface IQueryFunc<R extends QueryResultRow = any> {
	(sql: string, args?: any[]): Promise<QueryResult<R>>
}

type DataOrCustom<T> = T | (() => string);



type BasicWhere<V> = never
	| { in: V[] } | { nin: V[] }
	| { ne: V | null } | { gt: V } | { gte: V } | { lt: V } | { lte: V }
	| { like: string } | { nlike: string }

type ConditionOption<E extends BasicEntity> = { [P in BasicFieldName<E>]?: Exclude<E[P], null | undefined> | BasicWhere<Exclude<E[P], null | undefined>> | null }

/** 搜索选项映射关系 */
interface ISearchOptionMap {
	/** tsquery搜索 */
	tsquery: string | string[] | {
		/** 关键字 */
		keywords: string | string[]
		/** 是否进行绝对匹配，如果为true，则只匹配关键字，否则匹配关键字前缀，默认false */
		absolute?: boolean
		/** 匹配方式，默认or */
		method?: "or" | "and"
	}
	/** like搜索 */
	like: string | {
		/** 搜索关键字 */
		keywords: string
		/** 搜索方式， left表示左匹配，right表示右匹配，both表示两边匹配，默认both */
		method?: "left" | "right" | "both"
	}
}

interface IWhereFn<E extends BasicEntity> {
	/**
	 * 设置where条件
	 * @param option 条件选项
	 * @param sql 自定义sql
	 * @param args sql参数
	 */
	where(option: ConditionOption<E>): this
	where(sql: string, args?: any[]): this
	where<S extends string>(sql: S, args: { [P in SQLStringArgs<S>]: any }): this

	/**
	 * 添加搜索条件
	 * 
	 * 搜索条件会单独生成一段where语句，并与其他where条件使用AND连接
	 * 
	 * @param type 搜索类型
	 * @param field 搜索字段
	 * @param option 搜索选项
	 */
	search<T extends keyof ISearchOptionMap>(type: T, field: BasicFieldName<E>, option: ISearchOptionMap[T]): this
}

interface IJoinFn<E extends BasicEntity> {
	/**
	 * 进行JOIN操作
	 * @param name 字段名称
	 * @param callback Join后的具体操作
	 */
	join<N extends JoinaFieldName<E>>(name: N, callback?: (joinner: IJoinBuilder<JoinedEntity<E, N>>) => any): this

	/**
	 * 进行JOIN操作
	 * @param name 字段名称
	 * @param alias JOIN别名
	 * @param callback Join后的具体操作
	 */
	join<N extends JoinaFieldName<E>>(name: N, alias: string, callback?: (joinner: IJoinBuilder<JoinedEntity<E, N>>) => any): this
}

interface IFilterFn<E extends BasicEntity> {
	/**
	 * 查询字段过滤
	 * @param columns 要查询的字段，如果补传递则查询所有字段
	 */
	filter(...columns: Array<BasicFieldName<E> | { [P in BasicFieldName<E>]+?: string }>): this
	/**
	 * 排除字段
	 * @param columns 要排除的字段
	 */
	exclude(...columns: Array<BasicFieldName<E>>): this
	/**
	 * 忽略当前实体的字段
	 */
	ignore(): this
	/**
	 * 添加列
	 * @param column 列名称
	 * @param alias 别名
	 */
	add(column: BasicFieldName<E>, alias?: string): this

	/**
	 * 添加自定义列
	 * @param column 自定义列
	 * @param alias 别名
	 */
	add(column: () => string, alias: string): this
}

interface IReturningFn<E extends BasicEntity> {
	/**
	 * 设置返回字段
	 * @param columns 要返回的字段
	 */
	returning(...columns: Array<BasicFieldName<E>>): this
}

interface IGroupFn<E extends BasicEntity> {
	/**
	 * 分组
	 * @param name 分组名称
	 */
	group(...name: Array<(keyof E) | (() => string)>): this
}

interface IOrderByOption {
	/** 对nul字段的处理方式 */
	nulls?: "first" | "last"
	/** 是否将排序字段添加到内部的分组中 */
	group?: boolean
}

export interface ISelectBuilder<E extends BasicEntity> extends IWhereFn<E>, IFilterFn<E>, IJoinFn<E>, IGroupFn<E> {
	/**
	 * 设置查询偏移
	 * @param offset 查询偏移
	 */
	offset(offset: number): this

	/**
	 * 设置查询数量
	 * @param limit 查询数量
	 */
	limit(limit: number): this

	/**
	 * 排序
	 * @param name 排序名称
	 * @param sort 排序方式
	 */
	order(name: string, sort: "asc" | "desc", option?: IOrderByOption): this
	/**
	 * 排序
	 * @param name 排序名称
	 * @param sort 排序方式
	 */
	order(name: keyof E, sort: "asc" | "desc", option?: IOrderByOption): this

	// __sql__(count: boolean): string

	/** 查询数据 */
	find(): Promise<E[]>

	/** 查询一条数据 */
	findOne(): Promise<E | null>

	/** 仅查询数量 */
	count(): Promise<number>

	/** 查询数据和数量 */
	findAndCount(): Promise<[data: E[], count: number]>

	/** 仅执行sql语句 */
	query<T>(): Promise<T[]>;
}

export interface IInsertBuilder<E extends BasicEntity> extends IReturningFn<E> {
	/**
	 * 设置要插入的数据
	 * @param data 要插入的数据
	 */
	data<Name extends BasicFieldName<E>>(data: { [P in Name]: DataOrCustom<E[P]> } | Array<{ [P in Name]: DataOrCustom<E[P]> }>): this;

	/** 执行插入 */
	query(): Promise<E[]>
}

export interface IUpdateBuilder<E extends BasicEntity> extends IWhereFn<E>, IReturningFn<E> {
	/**
	 * 设置更新数据
	 * @param data 要更新的数据
	 */
	set(data: { [P in BasicFieldName<E>]?: DataOrCustom<E[P]> }): this

	/** 执行更新 */
	query(): Promise<E[]>
}

export interface IDeleteBuilder<E extends BasicEntity> extends IWhereFn<E>, IReturningFn<E> {
	/** 执行更新 */
	query(): Promise<E[]>
}

interface IJoinBuilder<E extends BasicEntity> extends IWhereFn<E>, IFilterFn<E>, IJoinFn<E> {
	/**
	 * 设置on条件
	 * @param option 条件选项
	 * @param sql 自定义sql
	 * @param args sql参数
	 */
	on(option: ConditionOption<E>): this
	on(sql: string, args?: any[]): this
	on<S extends string>(sql: S, args: { [P in StringArgs<S, "{:", "}">]: any }): this
}


type Real<T> = Exclude<T, null | undefined>
type JoinedEntity<E, N extends keyof E> = Real<E[N]> extends Array<infer R> ? (R extends BasicEntity ? R : never) : (Real<E[N]> extends BasicEntity ? Real<E[N]> : never);

type Constructor<E extends BasicEntity> = new (...args: any[]) => SQLBuilder<E>;

class SQLBuilderContext {
	#nameDict: Record<string, number> = {};
	#groups: string[] = [];
	#searchs: string[] = [];
	#hasManyJoin = false;

	public genName(name: string) {
		this.#nameDict[name] ??= 0;
		const idx = ++this.#nameDict[name];
		return `${name}${idx}`;
	}

	public addGroup(name: string) { this.#groups.push(name); }

	public get groups() { return this.#groups; }

	public addSearch(search: string) { this.#searchs.push(search); }

	public get searchs() { return this.#searchs; }

	public set hasManyJoin(val: boolean) { this.#hasManyJoin = val; }

	public get hasManyJoin() { return this.#hasManyJoin; }
}

class SQLBuilder<E extends BasicEntity> {
	#E: Class<E>
	#alias: string
	#ctx: SQLBuilderContext

	constructor(ctx: SQLBuilderContext, entity: Class<E>, alias: string | undefined) {
		this.#ctx = ctx;
		this.#E = entity;
		this.#alias = alias ?? this.tableName;
	}

	/** 实体类 */
	public get clazz() { return this.#E; }

	/** 上下文 */
	public get ctx() { return this.#ctx; }

	/** 实体配置 */
	public get config() { return getEntityConfig(this.#E); }

	/** 取表别名 */
	public get alias() { return this.#alias; }

	/** 表名称 */
	public get tableName() { return getTableName(this.#E); }

	/** 获取表名SQL */
	public get tableSQL() {
		const config = getEntityConfig(this.#E);
		if (config.schema) return escapeID(config.schema, config.table);
		return escapeID(config.table);
	}

	/** 字段遍历 */
	public eachField<R = unknown>(callback: (field: IEntityFieldConfig) => R): Array<R> {
		const fields = getEntityConfig(this.#E).fields;
		return fields.map(f => callback(f))
	}

	/** 获取字段配置 */
	public fieldc(prop: string) {
		return getField(this.#E, prop);
	}

	/** 获取join配置 */
	public joinc(prop: string) {
		return getJoin(this.#E, prop);
	}

	/** 获取主键字段 */
	public get primaries() {
		return getEntityConfig(this.#E).fields.filter(f => f.primary);
	}

	/** 获取主键字段名 */
	public get primaryNames() {
		return this.primaries.map(p => p.name);
	}

	public buildEntity(data: any) {
		const e = new this.#E();
		for (const field of this.config.fields) {
			const val = data[`${this.alias}__${field.prop}`];
			e[field.prop as keyof E] = (val === null || val === undefined) ? val : field.decode(val);
		}
		return e;
	}

	#checkPrimary<D>(dataItem: D) {
		return this.primaries.every(p => {
			const val = (dataItem as any)[`${this.alias}__${p.prop}`];
			return val !== null && val !== undefined;
		})
	}

	public __build__<D>(data: D[], joinners: Joinner<any>[]) {
		const map: Array<[E, D[]]> = [];

		//数据分组
		for (const item of data) {
			//空主键
			if (!this.#checkPrimary(item)) continue;
			let mapItem = map.findLast(([e, d]) => {
				return this.primaries.every(p => e[p.prop as keyof E] === item[`${this.alias}__${p.prop}` as keyof D] as any)
			});
			if (mapItem) mapItem[1].push(item);
			else map.push([this.buildEntity(item) as E, [item]]);
		}

		//子数据
		return map.map(([val, data]) => {
			for (const joinner of joinners) joinner.build(val, data);
			return val;
		});
	}
}


function WithFilter<E extends BasicEntity, B extends Constructor<E>>(Base: B) {
	return class extends Base implements IFilterFn<E> {
		#columns?: Array<[string, string]>;
		#customColumns: Array<[string, string]> = [];

		public filter(...columns: Array<BasicFieldName<E> | { [P in BasicFieldName<E>]+?: string }>) {
			this.#columns = [];
			if (!columns.length) columns = this.config.fields.map(f => f.prop as BasicFieldName<E>);
			for (const col of columns) {
				if (typeof col == "string") this.#columns.push([escapeID(this.alias, this.fieldc(col).name), escapeID(`${this.alias}__${col}`)]);
				else for (const [k, asName] of Object.entries(col)) {
					this.#columns.push([escapeID(this.alias, this.fieldc(k as string).name), asName as string]);
				}
			}
			return this;
		}

		public exclude(...columns: Array<BasicFieldName<E>>) {
			this.#columns = [];
			for (const col of this.config.fields) {
				if (columns.includes(col as any)) continue;
				this.#columns.push([escapeID(this.alias, col.name), escapeID(`${this.alias}__${col.prop}`)]);
			}
			return this;
		}

		public ignore() {
			this.#columns = [];
			return this;
		}

		public add(column: any, alias?: string) {
			if (typeof column == "string") this.#customColumns.push([escapeID(column), escapeID(alias ?? column)]);
			else if (typeof column == "function") this.#customColumns.push([column(), escapeID(alias!)]);
			return this;
		}

		/** 字段列表 */
		public get columns() {
			//如果没有则构建所有字段
			if (!this.#columns) this.filter(...this.eachField(f => f.prop as BasicFieldName<E>));
			//返回字段
			return [...this.#columns!, ...this.#customColumns];
		}
	}
}


/**
 * 构建条件语句，可用于where、having、on等语句
 * @param builder SQLBuilder
 * @param option 条件选项或条件语句
 * @param args 条件语句参数
 */
function buildCondition<E extends BasicEntity>(builder: SQLBuilder<E>, option: ConditionOption<any> | string, args?: any) {
	if (typeof option == "string") {
		if (typeof args === "undefined") return option;
		else return formatSQL(option, args);
	}
	else {
		const wheres: string[] = [];
		for (const [col, topVal] of Object.entries(option)) {
			const field = builder.fieldc(col);
			const fieldName = escapeID(builder.alias, field.name);
			if (topVal === null) wheres.push(`${fieldName} is null`);
			else if (typeof topVal == "object") {
				for (let [op, opv] of Object.entries(topVal as BasicWhere<any>) as [string, any][]) {
					switch (op) {
						case "in":
							wheres.push(`${fieldName} in (${(opv as any[]).map(v => field.encode(v)).join(",")})`);
							break;
						case "nin":
							wheres.push(`${fieldName} not in (${(opv as any[]).map(v => field.encode(v)).join(",")})`);
							break
						case "ne":
							if (opv === null) wheres.push(`${fieldName} is not null`);
							else wheres.push(`${fieldName}!=${field.encode(opv)}`);
							break;
						case "gt":
							wheres.push(`${fieldName}>${field.encode(opv)}`);
							break;
						case "gte":
							wheres.push(`${fieldName}>=${field.encode(opv)}`);
							break;
						case "lt":
							wheres.push(`${fieldName}<${field.encode(opv)}`);
							break;
						case "lte":
							wheres.push(`${fieldName}<=${field.encode(opv)}`);
							break;
						case "like":
							wheres.push(`${fieldName} like ${escapeValue(opv)}`);
							break;
						case "nlike":
							wheres.push(`${fieldName} note like ${escapeValue(opv)}`);
							break;
						default:
							break;
					}
				}
			}
			else if (typeof topVal !== "undefined") wheres.push(`${fieldName}=${field.encode(topVal)}`);
		}
		return wheres.join(" and ");
	}
}

/** Where 混入 */
function WithWhere<E extends BasicEntity, B extends Constructor<E>>(Base: B) {
	return class extends Base implements IWhereFn<E> {
		#wheres: string[] = [];

		public where(option: ConditionOption<any> | string, args?: any) {
			const where = buildCondition(this, option, args);
			if (where) this.#wheres.push(where);
			return this;
		}

		public search<T extends keyof ISearchOptionMap>(type: T, field: BasicFieldName<E>, option: ISearchOptionMap[T]) {
			const fieldName = escapeID(this.alias, field as string);
			switch (type) {
				case "tsquery": {
					const opt = option as ISearchOptionMap["tsquery"];
					let keywords: string[] = [];
					let method: "or" | "and" = "or";
					let absolute = false;

					const resolveKeyword = (kw: string) => {
						for (const s of kw.matchAll(/([a-zA-Z0-9\u4e00-\u9fa5]+)/g)) {
							if (keywords.includes(s[1])) continue;
							keywords.push(s[1]);
						}
					}

					//处理选项
					if (typeof opt === "string") resolveKeyword(opt);// keywords = opt.split(/\s+/).map(s => s.trim()).filter(s => s.length);
					else if (opt instanceof Array) keywords = opt;
					else {
						if (typeof opt.keywords === "string") resolveKeyword(opt.keywords) //keywords = opt.keywords.split(/\s+/).map(s => s.trim()).filter(s => s.length);
						else opt.keywords.forEach(kw => resolveKeyword(kw));
						if (opt.method) method = opt.method;
						if (typeof opt.absolute == "boolean") absolute = opt.absolute;
					}
					//生成tsquery
					if (keywords.length) {
						const sql = keywords.map(k => absolute ? k : `${k}:*`).join(method == "or" ? "|" : "&");
						this.ctx.addSearch(`${fieldName} @@ to_tsquery(${escapeValue(sql)})`);
					}
					break;
				}
				case "like": {
					const opt = option as ISearchOptionMap["like"];
					const keywords = ((typeof opt === "string") ? opt : opt.keywords).trim();
					if (keywords.length) {
						const method = (typeof opt === "object") ? (opt.method ?? "both") : "both";
						switch (method) {
							case "left":
								this.ctx.addSearch(`${fieldName} like ${escapeValue(`${keywords}%`)}`);
								break;
							case "right":
								this.ctx.addSearch(`${fieldName} like ${escapeValue(`%${keywords}`)}`);
								break;
							case "both":
								this.ctx.addSearch(`${fieldName} like ${escapeValue(`%${keywords}%`)}`);
								break;
							default:
								break;
						}
					}
					break;
				}
			}
			return this;
		}

		public get wheres() { return this.#wheres; }
	};
}

/** Join 混入 */
function WithJoin<E extends BasicEntity, B extends Constructor<E>>(Base: B) {
	return class extends Base implements IJoinFn<E> {
		#joinners: Joinner<any>[] = [];

		public join(name: string, alias?: any, callback?: any) {
			const joinConfig = this.joinc(name as string);
			const JoinEntity = joinConfig.entity();
			if (typeof callback == "undefined" && typeof alias == "function") {
				callback = alias;
				alias = undefined;
			}
			if (joinConfig.many) this.ctx.hasManyJoin = true;
			const joinner = new Joinner(this, name, this.alias, JoinEntity, joinConfig, alias ?? name);
			this.#joinners.push(joinner);
			if (callback) callback(joinner);
			return this;
		}

		public get joinners() { return this.#joinners; }
	}
}

/** Group by 混入 */
function WithGroup<E extends BasicEntity, B extends Constructor<E>>(Base: B) {
	return class extends Base implements IGroupFn<E> {

		public group(...name: Array<keyof E | (() => string)>) {
			for (const item of name) {
				if (typeof item == "function") this.ctx.addGroup(item());
				else if (typeof item == "string") this.ctx.addGroup(escapeID(this.alias, item));
			}
			return this;
		}
	}
}

/** Returning 混入 */
function WithReturning<E extends BasicEntity, B extends Constructor<E>>(Base: B) {
	return class extends Base implements IReturningFn<E> {
		#returning?: string[]

		public returning(...columns: Array<BasicFieldName<E>>) {
			if (!columns.length) columns = this.config.fields.map(c => c.name as BasicFieldName<E>);
			this.#returning = columns.map(c => `${escapeID(this.alias, c as string)} as ${escapeID(`${this.alias}__${c as string}`)}`);
			return this;
		}

		public get returnings() { return this.#returning; }

		public get returningsWithDefault() {
			return this.returnings ?? this.primaries.map(c => `${escapeID(this.alias, c.name)} as ${escapeID(`${this.alias}__${c.prop}`)}`);
		}

		public buildReturningData<D>(data: D[]) { return data.map(d => this.buildEntity(d)); }
	}
}

class Joinner<E extends BasicEntity> extends WithGroup(WithJoin(WithWhere(WithFilter(SQLBuilder))))<E> implements IJoinBuilder<E> {
	#joinName: string
	#joinConfig: IEntityJoinConfig
	#baseName: string
	#origin: SQLBuilder<any>
	#ons: string[] = [];

	public constructor(origin: SQLBuilder<any>, baseName: string, joinName: string, entity: Class<E>, config: IEntityJoinConfig, alias: string) {
		super(origin.ctx, entity, alias);
		this.#origin = origin;
		this.#baseName = baseName;
		this.#joinName = joinName;
		this.#joinConfig = config;
	}

	public on(option: ConditionOption<any> | string, args?: any) {
		const onSQL = buildCondition(this, option, args);
		if (onSQL) this.#ons.push(onSQL);
		return this;
	}

	public get __select_columns__(): Array<[name: string, alias: string]> {
		const result = [...this.columns];
		this.joinners.forEach(j => result.push(...j.__select_columns__));
		return result;
	}

	public get __wheres__(): string[] {
		const result = [...this.wheres];
		this.joinners.forEach(j => result.push(...j.__wheres__));
		return result;
	}

	public get __join_sql__() {
		const onSQL = [
			`${this.#joinConfig.builder(this.#joinName, this.alias)}`,
			...this.#ons,
		].join(" and ");
		return `left join ${this.tableSQL} ${escapeID(this.alias)} on ${onSQL}`;
	}

	public build<BE extends BasicEntity, D>(base: BE, data: D[]) {
		if (!this.columns.length) return;
		const parsed = this.__build__(data, this.joinners);
		if (this.#joinConfig.many) base[this.#baseName as keyof BE] = parsed as any;
		else base[this.#baseName as keyof BE] = (parsed[0] ?? null) as any;
	}
}

export class SelectBuilder<E extends BasicEntity> extends WithGroup(WithJoin(WithWhere(WithFilter(SQLBuilder))))<E> implements ISelectBuilder<E> {
	#offset?: number
	#limit?: number
	#orders?: Array<[name: string, order: "asc" | "desc", nulls: "first" | "last" | null, group: boolean]>
	#query: IQueryFunc<any>
	#fullJoinners?: Joinner<any>[]

	constructor(query: IQueryFunc<any>, entity: Class<E>, alias?: string) {
		super(new SQLBuilderContext(), entity, alias);
		this.#query = query;
	}

	public offset(offset: number) {
		this.#offset = offset;
		return this;
	}

	public limit(limit: number) {
		this.#limit = limit;
		return this;
	}

	public order(name: any, sort: "asc" | "desc", option?: IOrderByOption) {
		this.#orders ??= [];
		const col = this.config.fields.find(f => f.prop === name);
		if (col) this.#orders.push([escapeID(this.alias, col.name), sort, option?.nulls ?? null, option?.group ?? false]);
		else this.#orders.push([name, sort, option?.nulls ?? null, option?.group ?? false]);
		return this;
	}

	public async find() {
		const sql = this.__sql__(false);
		// console.log(sql);
		const result = await this.#query(sql);
		const ret = this.build(result.rows);
		return ret as E[];
	}

	async findOne() { return this.find().then(res => res[0] ?? null); }

	async count() {
		let sql: string;
		if (this.ctx.hasManyJoin) {
			sql = `select count(*) as __data_count__ from (\n\t${this.__inner_sql__(false, "\n\t")}\n)`;
		}
		else {
			sql = [
				`select count(*) as __data_count__`,
				`from ${this.tableSQL} ${escapeID(this.alias)}`,
				...this.__joins__,																			//join
				...this.__wheres__.length ? [`where ${this.__wheres__}`] : [],								//where条件
				...this.ctx.groups.length ? [`group by ${this.ctx.groups.join(",")}`] : [],					//分组
			].join("\n");
		}
		// console.log(sql);
		const res = await this.#query(sql);
		return parseInt(res.rows[0].__data_count__) as number
	}

	async findAndCount() {
		const res = await this.#query(this.__sql__(true));
		const count = res.rows[0]?.__data_count__ ?? 0;
		return [this.build(res.rows) as E[], parseInt(count)] as [E[], number];
	}

	async query<T>(): Promise<T[]> {
		const sql = this.__sql__(false);
		const res = await this.#query(sql);
		return res.rows;
	}

	public get __select_columns__() {
		const result = [...this.columns];
		this.joinners.forEach(j => result.push(...j.__select_columns__));
		return result.map(([c, a]) => `${c} as ${a}`)
	}

	#whereCache: string | undefined = undefined;
	public get __wheres__() {
		if (!this.#whereCache) {
			const result = [...this.wheres];
			this.joinners.forEach(j => result.push(...j.__wheres__));
			if (this.ctx.searchs.length) result.push("(" + this.ctx.searchs.join(" or ") + ")")
			this.#whereCache = result.join(" and ");
		}
		return this.#whereCache;
	}

	private getAllJoinners(joinners: Joinner<any>[]) {
		const result = [...joinners];
		joinners.forEach(j => result.push(...this.getAllJoinners(j.joinners)));
		return result;
	}

	public get __joins__() {
		this.#fullJoinners ??= this.getAllJoinners(this.joinners);
		return this.#fullJoinners.map(j => j.__join_sql__)
	}

	public __inner_sql__(count: boolean, join = "\n") {
		const groups = [...this.primaryNames.map(p => escapeID(this.alias, p))];
		this.#orders?.filter(([name, ord, nulls, group]) => group).forEach(([name, ord, nulls, group]) => groups.push(name));
		//内部的join
		const innerSQLItems = [
			`select ${escapeID(this.alias)}.*${count ? `, count(*) over () as __data_count__` : ''}`,			 					//字段列表
			`from ${this.tableSQL} ${escapeID(this.alias)}`,																		//表名
			...this.__joins__,																										//join
			...this.__wheres__.length ? [`where ${this.__wheres__}`] : [],															//where条件
			`group by ${groups.join(",")}`,											//内部查询分组
			...(count && this.#orders?.length) ? [`order by ${this.#orders.map(([name, ord, nulls]) => `${name} ${ord}${nulls ? ` nulls ${nulls}` : ""}`).join(",")}`] : [],		//排序
			...(typeof this.#offset == "number") ? [`offset ${this.#offset}`] : [],													//offset
			...(typeof this.#limit == "number") ? [`limit ${this.#limit}`] : [],													//limit
		];
		return innerSQLItems.join(join);
	}

	public __sql__(count: boolean) {
		if (this.ctx.hasManyJoin) {
			const groups = [...this.primaryNames.map(p => escapeID(this.alias, p))];
			this.#orders?.filter(([name, ord, nulls, group]) => group).forEach(([name, ord, nulls, group]) => groups.push(name));
			const buffer = [
				`select ${this.__select_columns__}${count ? `, ${escapeID(this.alias)}.__data_count__` : ''}`,			//字段列表
				`from (`,
				`\t${this.__inner_sql__(count, "\n\t")}`,
				`) ${escapeID(this.alias)}`,
				...this.__joins__,																						//join
				...this.__wheres__.length ? [`where ${this.__wheres__}`] : [],											//where条件
				...this.ctx.groups.length ? [`group by ${this.ctx.groups.join(",")}`] : [],								//分组
				...this.#orders ? [`order by ${this.#orders.map(([name, ord, nulls]) => `${name} ${ord}${nulls ? ` nulls ${nulls}` : ""}`).join(", ")}`] : [],	//排序
			];
			// console.log(buffer.join("\n"));
			return buffer.join("\n");
		}
		else {
			const buffer = [
				`select ${this.__select_columns__}${count ? `,count(*) over () as __data_count__` : ''}`,				//字段列表
				`from ${this.tableSQL} ${escapeID(this.alias)}`,														//表名
				...this.__joins__,																						//join
				...this.__wheres__.length ? [`where ${this.__wheres__}`] : [],											//where条件
				...this.ctx.groups.length ? [`group by ${this.ctx.groups.join(",")}`] : [],								//分组
				...this.#orders ? [`order by ${this.#orders.map(([name, ord, nulls]) => `${name} ${ord}${nulls ? ` nulls ${nulls}` : ""}`).join(", ")}`] : [],	//排序
				...(typeof this.#offset == "number") ? [`offset ${this.#offset}`] : [],									//offset
				...(typeof this.#limit == "number") ? [`limit ${this.#limit}`] : [],									//limit
			];
			// console.log(buffer.join("\n"));
			return buffer.join("\n");
		}
	}

	public build<D>(data: D[]) {
		return this.__build__(data, this.joinners);
	}
}

export class InsertBuilder<E extends BasicEntity> extends WithReturning(SQLBuilder)<E> implements IInsertBuilder<E> {
	#query: IQueryFunc<any>
	#values: string[] = [];
	#columns: string[] = [];

	public constructor(query: IQueryFunc<any>, entity: Class<E>) {
		super(new SQLBuilderContext(), entity, undefined);
		this.#query = query;
	}

	public data<Name extends BasicFieldName<E>>(data: { [P in Name]: DataOrCustom<E[P]> } | Array<{ [P in Name]: DataOrCustom<E[P]> }>) {
		data = (data instanceof Array) ? data : [data];
		this.#values = [];
		this.#columns = [];
		let columnsSet = false;

		for (const item of data) {
			const buffer: (string | null)[] = [];
			for (const col of this.config.fields) {
				if (col.virtual) continue;
				if (!(col.prop in item)) continue;
				const val = item[col.prop as keyof typeof item];
				if (val === null || val === undefined) buffer.push("null");
				else if (typeof val === "function") buffer.push(val());
				else buffer.push(col.encode(val));
				if (!columnsSet) this.#columns.push(escapeID(col.name));
			}
			this.#values.push(`(${buffer.join(",")})`);
			columnsSet = true;
		}

		return this;
	}

	public async query() {
		if (!this.#values.length) return [];
		//构建SQL
		const sqlItems = [
			`insert into ${this.tableSQL} as ${escapeID(this.alias)} (${this.#columns.join(",")}) values`,
			this.#values.join(",\n"),
			`returning ${this.returningsWithDefault}`,
		]
		const sql = sqlItems.join("\n");

		//执行并返回数据
		const res = await this.#query(sql);
		return this.buildReturningData(res.rows) as E[];
	}
}

export class UpdateBuilder<E extends BasicEntity> extends WithReturning(WithWhere(SQLBuilder))<E> implements IUpdateBuilder<E> {
	#query: IQueryFunc<any>
	#updates?: string[]

	public constructor(query: IQueryFunc<any>, entity: Class<E>) {
		super(new SQLBuilderContext(), entity, undefined);
		this.#query = query;
	}

	public set(data: { [P in BasicFieldName<E>]?: DataOrCustom<E[P]> }) {
		this.#updates = []
		for (const col of this.config.fields) {
			if (col.virtual) continue;
			const val = data[col.prop as keyof typeof data];
			if (val === undefined) continue;
			else if (val === null) this.#updates.push(`${escapeID(col.name)}=null`);
			else if (typeof val === "function") this.#updates.push(`${escapeID(col.name)}=${val()}`);
			else this.#updates.push(`${escapeID(col.name)}=${col.encode(val)}`);
		}
		return this;
	}

	public async query() {
		if (!this.#updates?.length) return [];
		//构建SQL
		const wheres = this.wheres;
		const sqlItems = [
			`update ${this.tableSQL} ${escapeID(this.alias)}`,
			`set ${this.#updates.join(",")}`,
			...wheres.length ? [`where ${wheres.join(" and ")}`] : [],
			`returning ${this.returningsWithDefault}`,
		]
		const sql = sqlItems.join("\n");

		//执行并构建返回数据
		const res = await this.#query(sql);
		return this.buildReturningData(res.rows) as E[];
	}

}

export class DeleteBuilder<E extends BasicEntity> extends WithReturning(WithWhere(SQLBuilder))<E> implements IDeleteBuilder<E> {
	#query: IQueryFunc<any>

	public constructor(query: IQueryFunc<any>, entity: Class<E>) {
		super(new SQLBuilderContext(), entity, undefined);
		this.#query = query;
	}

	public async query() {
		//构建SQL
		const wheres = this.wheres;
		const sqlItems = [
			`delete from ${this.tableSQL} ${escapeID(this.alias)}`,
			...wheres.length ? [`where ${wheres.join(" and ")}`] : [],
			`returning ${this.returningsWithDefault}`,
		]
		const sql = sqlItems.join("\n");

		//执行并构建返回数据
		const res = await this.#query(sql);
		return this.buildReturningData(res.rows) as E[];
	}
}
