import type { BasicEntity } from "./entity";

export type Class<T> = new (...args: any[]) => T;

export type ArrayType<T> = T extends Array<infer U> ? U : never;
export type ArrayTypeSelf<T> = T extends Array<infer U> ? U : T;

export type Values<T> = T[keyof T];

export namespace Decorators {
	export type ClassDecorator<T> = (target: Class<T>) => any;
	export type PropDecorator<T> = (target: T, propertyKey: string) => any;
	export type MethodDecorator<T> = (target: T, propertyKey: string, descriptor: PropertyDescriptor) => any;
	export type ParamDecorator<T> = (target: T, propertyKey: string, parameterIndex: number) => any;
}

declare global {

	/** 取基础类型的字段名 */
	type BasicFieldName<T extends BasicEntity> = Values<{ [P in keyof T]:
		//忽略never
		Exclude<T[P], null | undefined> extends never ? never : (
			//忽略函数
			Exclude<T[P], null | undefined> extends Function ? never : (
				//忽略实体
				ArrayTypeSelf<Exclude<T[P], null | undefined>> extends BasicEntity ? never : (
					P
				)
			)
		)
	}>;
}


/** 可以用作ID的字段名称 */
export type IDLikeFieldName<T extends BasicEntity> = Values<{ [P in keyof T]:
	Exclude<T[P], null | undefined> extends never ? never : (
		string extends Exclude<T[P], null | undefined> ? P : (
			number extends Exclude<T[P], null | undefined> ? P : never
		)
	)
}>;


/** 获取可用于ManyJoin的字段名 */
export type ManyJoinFieldName<T extends BasicEntity> = Values<{ [P in keyof T]:
	//忽略never
	ArrayType<Exclude<T[P], null | undefined>> extends never ? never : (
		//只取Array实体
		ArrayType<Exclude<T[P], null | undefined>> extends BasicEntity ? P : never
	)
}>;

/** 获取可以用户OneJoin的字段名 */
export type OneJoinFieldName<T extends BasicEntity> = Values<{ [P in keyof T]:
	//忽略never
	Exclude<T[P], null | undefined> extends never ? never : (
		//只取Array实体
		Exclude<T[P], null | undefined> extends BasicEntity ? P : never
	)
}>;

/** 可以Join的字段名 */
export type JoinaFieldName<T extends BasicEntity> = OneJoinFieldName<T> | ManyJoinFieldName<T>;

/** 
 * SQL字符串参数 
 * 
 * @param Str SQL字符串
 * @param P 前分割
 * @param S 后分割
 */
export type StringArgs<Str extends string, P extends string, S extends string> = Str extends `${string}${P}${infer R}${S}${infer T}` ? (R | StringArgs<T, P, S>) : never;

/** 
 * SQL字符串参数 
 * 
 * @param Str SQL字符串
 */
export type SQLStringArgs<Str extends string> = StringArgs<Str, "{:", "}">;
